#------------------------------------------------------------------
# VT50-2000-2016-84-28-4
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
# $Header: /project/isv-agent/VProAgents/SystemD/SystemD.pm,v 1.20 2015/06/18 11:30:42 avaidya Exp $ 
#
# $Author: avaidya $
# $Date: 2015/06/18 11:30:42 $
# $Revision: 1.20 $
#
# Provides:                            SystemD
# Author:                              avaidya
# crAgent Build:                       1.88 2008/10/13 08:28:02
# crAgent Run Date:                    13 Oct, 2008
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;
use warnings;

#----------------------
# Declare the package..
#----------------------
package SystemD;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );
use POSIX qw ( strftime );
use File::Path qw ( mkpath );
use File::stat;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vcs(|one))$/;

if ( $sPrefix =~ /vcsone/ ) {
   $sPrefix = "ha";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SystemD.pm,v 1.20 2015/06/18 11:30:42 avaidya Exp $::ETVHeader";
Version::TellVersion ( $sVer );
my $SYSTEMCTL = "/usr/bin/systemctl";
my $GREP = "/bin/grep";

#-----------------
# Define globals..
#-----------------

#--------------------------------
# Define SystemD subroutines..
#--------------------------------

=pod
 -------------------------------------------------------------------------
 Function : int DetectSystemd ();

 Description : Retrieves whether current operating system is linux(RHEL7/SLES12 or higher). 
 
 Parameters  :

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub DetectSystemd {
  #----------------
  # Get Arguments..
  #----------------

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20900, "Subroutine <%s> called .",
    $sMyName);

  #-------------------------
  # Define local variables..
  #-------------------------
  my ($cmd);
  my ($is_rhel7, $is_sles12)= ( 1, 1);
  my ($pid_one, $tty, $time, @psargs);
  my ($processname);
  my ($SYSTEM) = uname();
  my $is_systemd = 0;

  if ($SYSTEM eq "Linux") {
    $is_rhel7 = system("$GREP -qs \'release 7.*\' /etc/redhat-release");
    $is_sles12 = system("$GREP -qs \'SUSE.*12\' /etc/SuSE-release");
    
    # Systemd is supported in case of RHEL7 and SLES12 only
    if( !$is_rhel7 || !$is_sles12){    
      $cmd = "/bin/ps -e";
      unless (open(PS, "$cmd | $GREP \" 1 ?\"|")) {
      	return 0;
    }
# output of ps command
#     1 ?        00:09:49 systemd
    while (my $line = <PS>) {
      chomp($line);
      $line =~ s/^\s*(.*?)\s*$/$1/;
      ($pid_one, $tty, $time, @psargs) = split(/\s+/, $line);
      ($processname) = join(' ', @psargs);

      if ( ( 1 == $pid_one ) && ($processname =~ /^systemd\b/) ) {
        $is_systemd = 1;
        last;
      }
    }
       
    close(PS);
    } #if (!$is_rhel7 || !is_sles12)
  } #if ($SYSTEM eq "Linux")
  return $is_systemd;

}

=pod
 -------------------------------------------------------------------------
 Function : int CreateSystemd ();

 Description : Creates systemd service file for each resource . 
 
 Parameters  :
 ResName       Resource name 
 User          user of the command for which service needs to be created
 Command       Command to be specified in the service file
 CommandArgs   Arguments of the command to be specified in the service file
 EnvFile       Enivironment File 
 TimeOutSec    timeout  needs to be passed to service file
 sCommandName  Command for which service needs to be created
 sHideString   String in CommandArgs to hide
 FromDir       Directory from where command to be executed

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreateSystemd {
  #----------------
  # Get Arguments..
  #----------------

  my($ResName, $User, $sCommand, $sCommandArgs, $sEnvFile, $TimeOutSec, $sCommandName,
    $sHideString,$sFromDir) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20901, "Subroutine <%s> called with:
  ResName           [%s]
  User              [%s]
  Command           [%s]
  CommandArgs       [%s]
  EnvFile           [%s]
  TimeOutSec        [%s]
  CommandName       [%s]
  FromDir           [%s]",
  $sMyName, $ResName, $User, $sCommand, $sCommandArgs, $sEnvFile, $TimeOutSec, $sCommandName,$sFromDir);

  #-------------------------
  # Define local variables..
  #-------------------------
  my $service_file = "/etc/systemd/system/vcs-$ResName.service";
  if ( defined $sCommandName ) {
    $service_file = "/etc/systemd/system/vcs-$ResName-$sCommandName.service";
  }
  my $ret_val = 0;#failure
  my $iExitCode = 0;
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20144,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }    

    my $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    my $sDeCrPasswd = readpipe ( "$sCmdDeCr $sHideString" ); # Get decrypted..

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20145, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/\'$sHideString\'/$sDeCrPasswd/;
    $sDeCrPasswd  =~ s/\$/./g;
  }


  if ( sysopen ( FILE, $service_file, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC, 0770 ) ) {

    printf FILE "#$service_file

[Unit]
Description=Veritas $sResourceType service file

[Service]
Type=forking
Restart=no
KillMode=none
SendSIGKILL=no
EnvironmentFile=$sEnvFile";
if ( defined $sFromDir ) {  
printf FILE "
WorkingDirectory=$sFromDir
RootDirectory=/";
}
printf FILE "
ExecStart=$sCommand $sCommandArgs
TimeoutStartSec=$TimeOutSec
User=$User

[Install]
WantedBy=multi-user.target";

	$ret_val = 1;
  }
  else
  {
    &logError ( 20902, "<%s> Failed to create the systemd service file [%s] ", 
    $sMyName,$service_file);
    $ret_val = 0;
  }
  close(FILE);
  return $ret_val;

}
=pod
 -------------------------------------------------------------------------
 Function : int RunSystemd ();

 Description : Runs systemd service file for each resource . 
 
 Parameters  :
 ResName       Resource name 
 sCommandName  Command for which service needs to be created
 sHideString   String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 FAILURE
           1 SUCCESS
 -------------------------------------------------------------------------

=cut

sub RunSystemd {
  #----------------
  # Get Arguments..
  #----------------

  my ( $ResName , $sCommandName,$sHideString) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20903, "Subroutine <%s> called with:
  ResName         [%s]
  CommandName     [%s].",
  $sMyName, $ResName, $sCommandName);

  #-------------------------
  # Define local variables..
  #-------------------------
  my $Command = "$SYSTEMCTL daemon-reload";
  &logTrace ( 20904, "<%s> Executing command [%s].",$sMyName,$Command);
  my $iExitCode = system($Command) ;
  my $iReturnCode = 1;
  my $serviceName = "vcs-$ResName";
  if ( defined $sCommandName ) {
    $serviceName = "vcs-$ResName-$sCommandName";
  }

  if ( $iExitCode == 0 ) {
    &logTrace ( 20905, "<%s> Cmd [%s] executed successfully",$sMyName, $Command);

    $Command = "$SYSTEMCTL disable $serviceName";
    &logTrace ( 20906, "<%s> Executing command [%s].",$sMyName,$Command);
    $iExitCode = system($Command) ;

    if ( $iExitCode == 0 ) {
      &logTrace ( 20907, "<%s> Cmd [%s] executed successfully",$sMyName, $Command);

      $Command = "$SYSTEMCTL start $serviceName";
      &logTrace ( 20908, "<%s> Executing command [%s].",$sMyName,$Command);
      $iExitCode = system($Command) ;

      if ( $iExitCode != 0 ) {
        &logError ( 20909, "<%s> Cmd [%s] is not executed successfully", $sMyName,$Command);
        $iReturnCode = 0;
      } else {
        &logTrace ( 20912, "<%s> Cmd [%s] executed successfully",$sMyName, $Command);
      }

      $Command = "$SYSTEMCTL -l status $serviceName";
      &logTrace ( 20910, "<%s> Executing command [%s].",$sMyName,$Command);

      if ( not defined $sHideString ) {
        my $outMessage = `$SYSTEMCTL -l status $serviceName 2>&1`;
        &logTrace ( 20911, "<%s> outMessage is [%s]",$sMyName, $outMessage);
      }
    } else {
      &logError ( 20913, "<%s> Cmd [%s] is not executed successfully",$sMyName, $Command);
      $iReturnCode = 0;
    }
 
  } else {
    &logError ( 20914, "<%s> Cmd [%s] is not executed successfully",$sMyName, $Command);
    $iReturnCode = 0;
  }

  return $iReturnCode;
}

=pod
 -------------------------------------------------------------------------
 Function : int RunSystemdWithEnv ();

 Description : Creates and Runs systemd service file for each resource . 
 
 Parameters  :
 User          user of the command for which service needs to be created
 Command       Command to be specified in the service file
 CommandArgs   Arguments of the command to be specified in the service file
 EnvFile       EnvFile which needs to be passed to service file
 TimeOutSec    timeout  needs to be passed to service file
 sCommandName  Command for which service needs to be created
 sHideString   String in CommandArgs to hide
 FromDir       Directory from where command to be executed

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RunSystemdWithEnv {
  #----------------
  # Get Arguments..
  #----------------

  my($User, $sCommand, $sCommandArgs, $sEnvFile, $iTimeOutSec, $sCommandName,$sHideString,$sFromDir) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20915, "Subroutine <%s> called with:
  User              [%s]
  Command           [%s]
  CommandArgs       [%s]
  EnvFile           [%s]
  TimeOutSec        [%s]
  CommandName       [%s]
  FromDir           [%s]",
  $sMyName, $User, $sCommand, $sCommandArgs, $sEnvFile, $iTimeOutSec, $sCommandName,$sFromDir);

  #-------------------------
  # Define local variables..
  #-------------------------
  &logTrace ( 20916,  "<%s> SYSTEMD is detected",$sMyName);
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; 

  my $ResName = $ENV { VCS_LOG_RESOURCE_NAME };
  $iReturnCode = CreateSystemd($ResName,$User,$sCommand,$sCommandArgs,$sEnvFile, $iTimeOutSec, $sCommandName,$sHideString,$sFromDir);
  if($iReturnCode) {
    &logTrace ( 20917,  "<%s> CreateSystemd was successful",$sMyName);
    $iReturnCode = RunSystemd($ResName,$sCommandName,$sHideString);
    if($iReturnCode) {
      &logTrace ( 20918, "<%s> RunSystemd was successful",$sMyName);
    } else {
      &logError ( 20919, "<%s> RunSystemd was not successful",$sMyName);
    }
  } else {
    &logError ( 20920, "<%s> CreateSystemd was not successful",$sMyName);
  }
  
  return $iReturnCode;
}


=pod
 -------------------------------------------------------------------------
 Function : int CreateCommandInBG ();

 Description : Creates commmand and its command arguments which will be executed in background.
 
 Parameters  :
 EnvFile         EnvFile which needs to be passed to service file
 Command         Command to be specified in the service file
 CommandArgs     Arguments of the command to be specified in the service file
 User            user of the command for which service needs to be created
 CommandOutFile  Output file where output from the command is stored

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub CreateCommandInBG
{
  #--------------
  # Get Arguments
  #--------------
  my ( $sEnvFile,
       $rsCommand,
       $rsCommandArgs,
       $sUser,
       $sCommandOutFile ) = @_ ;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20921,
  "Subroutine <%s> called with:
  EnvFile               [%s]
  Command               [%s]
  CommandArgs           [%s]
  User                  [%s] 
  CommandOutFile        [%s] ",
  $sMyName, $sEnvFile, $$rsCommand,
  $$rsCommandArgs, $sUser, $sCommandOutFile);

  #--------------------------
  # Establish local variables
  #--------------------------
  my $sCmdToRun    = "";
  my $sCommandLine = "";
  my $sRedirection = "";

  my ( $sName,
       $sPassword,
       $iUid,
       $iGid,
       $iQuota,
       $sComment,
       $sGcos,
       $sHome,
       $sUserShell,
       $sExpire     ) = getpwnam( $sUser );

  #-----------------------------
  # Confirm existance of EnvFile
  #-----------------------------
  $sRedirection = "> $sCommandOutFile 2>&1 &";
  if( $sEnvFile eq "" )
  {
    &logTrace (20924, "<%s> Environment file not set", $sMyName );
  }
  elsif( ! -r $sEnvFile )
  {
    &logError ( 20925, "<%s> Environment file [%s] does not exist", $sMyName , $sEnvFile );
    return 0;
  }
  else
  {
    if( $sUserShell =~ /(.*)\/bin\/.*csh$/ )
    {
      &logTrace (20926, "<%s> User [%s] has csh derived native shell [%s]", $sMyName , $sUser, $sUserShell );
      #---------------------------------------------------------------
      # Shell is a csh varient, so we should use the 'source' command.
      #---------------------------------------------------------------
      $sRedirection = "& > $sCommandOutFile 2>&1";
    }
    else
    {
      &logTrace ( 20927, "<%s> User [%s] has sh derived native shell [%s]", $sMyName , $sUser, $sUserShell );
      #---------------------------------------------------------
      # Shell is one of the non-C type shells. Use '.' to source 
      # in the env file.
      #---------------------------------------------------------
      $sRedirection = "> $sCommandOutFile 2>&1 &";
    }
  }

  #--------------------------------
  # Confirm existance of Exectuable
  #--------------------------------
  if( ! -f $$rsCommand )
  {
    &logError ( 20928, "<%s> File [%s] does not exist or is not a file", $sMyName , $$rsCommand ) ;
    return 0;
  }
  elsif( ! -x $$rsCommand )
  {
    &logError ( 20929, "<%s> File [%s] is not executable", $sMyName , $$rsCommand ) ;
    return 0;
  }

  $$rsCommandArgs = " " . "$$rsCommandArgs $sRedirection"; 
  &logTrace ( 20930, "<%s> Going to run command line [%s], as user [%s]", $sMyName, $$rsCommand, $sUser );
  return 1;
}

=pod
 -------------------------------------------------------------------------
 Function : int RunSystemdWithEnvInBG ();

 Description : Creates and Runs systemd service file for each resource. The process for each 
               resource will be executed in background.
 
 Parameters  :
 EnvFile         EnvFile which needs to be passed to service file
 Command         Command to be specified in the service file
 CommandArgs     Arguments of the command to be specified in the service file
 User            user of the command for which service needs to be created
 CommandOutFile  Output file where output from the command is stored
 TimeOutSec      timeout  needs to be passed to service file
 RunCmdFrom      directory from where command is executed
 CommandName     Command for which service needs to be created
 HideString      String in CommandArgs to hide
 FromDir         Directory from where command to be executed 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub RunSystemdWithEnvInBG
{
  #--------------
  # Get Arguments
  #--------------
  my ( $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $sUser,
       $sCommandOutFile,
       $iTimeoutSeconds,
       $sCommandName,
       $sHideString,
       $sFromDir       ) = @_ ;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20921,
  "Subroutine <%s> called with:
  EnvFile               [%s]
  Command               [%s]
  CommandArgs           [%s]
  User                  [%s]
  CommandOutFile        [%s]
  TimeoutSeconds        [%s]
  CommandName           [%s]
  FromDir               [%s]",
  $sMyName, $sEnvFile, $sCommand,
  $sCommandArgs, $sUser, $sCommandOutFile, $iTimeoutSeconds, $sCommandName,$sFromDir );

  #--------------------------
  # Establish local variables
  #--------------------------
  my $sCmdToRun    = "";
  my $iReturnCode  = 1 ;

  if( ! defined $sUser || $sUser eq "" )
  {
    $sUser = "root";
  }

  if( $iTimeoutSeconds < 1 )
  {
    &logTrace ( 20932, "<%s> Passed in timeout was less than 1 second, defaulting it to 15 seconds" , $sMyName);
    $iTimeoutSeconds = 15;
  }
  
  if( $iReturnCode = CreateCommandInBG($sEnvFile,\$sCommand, \$sCommandArgs,$sUser,$sCommandOutFile) ){
    &logTrace ( 20933, "<%s> CreateCommandInBG was successful", $sMyName);
  }
  else {
    &logError ( 20934, "<%s> CreateCommandInBG was not successful", $sMyName);
    return $iReturnCode;
  }
   
  if( $iReturnCode = RunSystemdWithEnv($sUser, $sCommand, $sCommandArgs, $sEnvFile, $iTimeoutSeconds,$sCommandName, $sHideString,$sFromDir) ) {
    &logTrace ( 20935, "<%s> RunSystemdWithEnv was successful", $sMyName);
  }
  else {
    &logError ( 20936, "<%s> RunSystemdWithEnv was not successful", $sMyName);
  }
  return $iReturnCode;
}

1; 



